#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import networkx as nx
import numpy as np
from graph_nets.utils_np import graphs_tuple_to_networkxs

from kglib.kgcn.learn.learn import KGCNLearner
from kglib.kgcn.models.core import softmax, KGCN
from kglib.kgcn.models.embedding import ThingEmbedder, RoleEmbedder
from kglib.kgcn.pipeline.encode import encode_types, create_input_graph, create_target_graph, encode_values
from kglib.kgcn.pipeline.utils import apply_logits_to_graphs, duplicate_edges_in_reverse
from kglib.kgcn.plot.plotting import plot_across_training, plot_predictions
from kglib.utils.graph.iterate import multidigraph_node_data_iterator, multidigraph_data_iterator, \
    multidigraph_edge_data_iterator


def pipeline(graphs,
             tr_ge_split,
             node_types,
             edge_types,
             num_processing_steps_tr=10,
             num_processing_steps_ge=10,
             num_training_iterations=10000,
             continuous_attributes=None,
             categorical_attributes=None,
             type_embedding_dim=5,
             attr_embedding_dim=6,
             edge_output_size=3,
             node_output_size=3,
             output_dir=None):

    ############################################################
    # Manipulate the graph data
    ############################################################

    # Encode attribute values
    graphs = [encode_values(graph, categorical_attributes, continuous_attributes) for graph in graphs]

    indexed_graphs = [nx.convert_node_labels_to_integers(graph, label_attribute='concept') for graph in graphs]
    graphs = [duplicate_edges_in_reverse(graph) for graph in indexed_graphs]

    graphs = [encode_types(graph, multidigraph_node_data_iterator, node_types) for graph in graphs]
    graphs = [encode_types(graph, multidigraph_edge_data_iterator, edge_types) for graph in graphs]

    input_graphs = [create_input_graph(graph) for graph in graphs]
    target_graphs = [create_target_graph(graph) for graph in graphs]

    tr_input_graphs = input_graphs[:tr_ge_split]
    tr_target_graphs = target_graphs[:tr_ge_split]
    ge_input_graphs = input_graphs[tr_ge_split:]
    ge_target_graphs = target_graphs[tr_ge_split:]

    ############################################################
    # Build and run the KGCN
    ############################################################

    thing_embedder = ThingEmbedder(node_types, type_embedding_dim, attr_embedding_dim, categorical_attributes,
                                   continuous_attributes)

    role_embedder = RoleEmbedder(len(edge_types), type_embedding_dim)

    kgcn = KGCN(thing_embedder,
                role_embedder,
                edge_output_size=edge_output_size,
                node_output_size=node_output_size)

    learner = KGCNLearner(kgcn,
                          num_processing_steps_tr=num_processing_steps_tr,
                          num_processing_steps_ge=num_processing_steps_ge)

    train_values, test_values, tr_info = learner(tr_input_graphs,
                                                 tr_target_graphs,
                                                 ge_input_graphs,
                                                 ge_target_graphs,
                                                 num_training_iterations=num_training_iterations,
                                                 log_dir=output_dir)

    plot_across_training(*tr_info, output_file=f'{output_dir}learning.png')
    plot_predictions(graphs[tr_ge_split:], test_values, num_processing_steps_ge, output_file=f'{output_dir}graph.png')

    logit_graphs = graphs_tuple_to_networkxs(test_values["outputs"][-1])

    indexed_ge_graphs = indexed_graphs[tr_ge_split:]
    ge_graphs = [apply_logits_to_graphs(graph, logit_graph) for graph, logit_graph in
                 zip(indexed_ge_graphs, logit_graphs)]

    for ge_graph in ge_graphs:
        for data in multidigraph_data_iterator(ge_graph):
            data['probabilities'] = softmax(data['logits'])
            data['prediction'] = int(np.argmax(data['probabilities']))

    _, _, _, _, _, solveds_tr, solveds_ge = tr_info
    return ge_graphs, solveds_tr, solveds_ge



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
import warnings
from functools import reduce

import networkx as nx

from kglib.utils.grakn.object.thing import build_thing
from kglib.utils.graph.thing.concept_dict_to_graph import concept_dict_to_graph


def concept_dict_from_concept_map(concept_map):
    """
    Given a concept map, build a dictionary of the variables present and the concepts they refer to, locally storing any
    information required about those concepts.

    Args:
        concept_map: A dict of Concepts provided by Grakn keyed by query variables

    Returns:
        A dictionary of concepts keyed by query variables
    """
    return {variable: build_thing(grakn_concept) for variable, grakn_concept in concept_map.map().items()}


def combine_2_graphs(graph1, graph2):
    """
    Combine two graphs into one. Do this by recognising common nodes between the two.

    Args:
        graph1: Graph to compare
        graph2: Graph to compare

    Returns:
        Combined graph
    """

    for node, data in graph1.nodes(data=True):
        if graph2.has_node(node):
            data2 = graph2.nodes[node]
            if data2 != data:
                raise ValueError((f'Found non-matching node properties for node {node} '
                                  f'between graphs {graph1} and {graph2}:\n'
                                  f'In graph {graph1}: {data}\n'
                                  f'In graph {graph2}: {data2}'))

    for sender, receiver, keys, data in graph1.edges(data=True, keys=True):
        if graph2.has_edge(sender, receiver, keys):
            data2 = graph2.edges[sender, receiver, keys]
            if data2 != data:
                raise ValueError((f'Found non-matching edge properties for edge {sender, receiver, keys} '
                                  f'between graphs {graph1} and {graph2}:\n'
                                  f'In graph {graph1}: {data}\n'
                                  f'In graph {graph2}: {data2}'))

    return nx.compose(graph1, graph2)


def combine_n_graphs(graphs_list):
    """
    Combine N graphs into one. Do this by recognising common nodes between the two.

    Args:
        graphs_list: List of graphs to combine

    Returns:
        Combined graph
    """
    return reduce(lambda x, y: combine_2_graphs(x, y), graphs_list)


def build_graph_from_queries(query_sampler_variable_graph_tuples, grakn_transaction,
                             concept_dict_converter=concept_dict_to_graph, infer=True):
    """
    Builds a graph of Things, interconnected by roles (and *has*), from a set of queries and graphs representing those
    queries (variable graphs)of those queries, over a Grakn transaction

    Args:
        infer: whether to use Grakn's inference engine
        query_sampler_variable_graph_tuples: A list of tuples, each tuple containing a query, a sampling function,
            and a variable_graph
        grakn_transaction: A Grakn transaction
        concept_dict_converter: The function to use to convert from concept_dicts to a Grakn model. This could be
            a typical model or a mathematical model

    Returns:
        A networkx graph
    """

    query_concept_graphs = []

    for query, sampler, variable_graph in query_sampler_variable_graph_tuples:

        concept_maps = sampler(grakn_transaction.query(query, infer=infer))

        concept_dicts = [concept_dict_from_concept_map(concept_map) for concept_map in concept_maps]

        answer_concept_graphs = []
        for concept_dict in concept_dicts:
            try:
                answer_concept_graphs.append(concept_dict_converter(concept_dict, variable_graph))
            except ValueError as e:
                raise ValueError(str(e) + f'Encountered processing query:\n \"{query}\"')

        if len(answer_concept_graphs) > 1:
            query_concept_graph = combine_n_graphs(answer_concept_graphs)
            query_concept_graphs.append(query_concept_graph)
        else:
            if len(answer_concept_graphs) > 0:
                query_concept_graphs.append(answer_concept_graphs[0])
            else:
                warnings.warn(f'There were no results for query: \n\"{query}\"\nand so nothing will be added to the '
                              f'graph for this query')

    if len(query_concept_graphs) == 0:
        # Raise exception when none of the queries returned any results
        raise RuntimeError(f'The graph from queries: {[query_sampler_variable_graph_tuple[0] for query_sampler_variable_graph_tuple in query_sampler_variable_graph_tuples]}\n'
                           f'could not be created, since none of these queries returned results')

    concept_graph = combine_n_graphs(query_concept_graphs)
    return concept_graph



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: session/Session.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from grakn_protocol.session import Concept_pb2 as session_dot_Concept__pb2
from grakn_protocol.session import Answer_pb2 as session_dot_Answer__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='session/Session.proto',
  package='session',
  syntax='proto3',
  serialized_options=_b('\n\026grakn.protocol.sessionB\014SessionProto'),
  serialized_pb=_b('\n\x15session/Session.proto\x12\x07session\x1a\x15session/Concept.proto\x1a\x14session/Answer.proto\"\xd1\x03\n\x07Session\x1a\x8a\x02\n\x04Open\x1a\xdc\x01\n\x03Req\x12\x44\n\x08metadata\x18\xe8\x07 \x03(\x0b\x32\'.session.Session.Open.Req.MetadataEntryR\x08metadata\x12\x1a\n\x08username\x18\x01 \x01(\tR\x08username\x12\x1a\n\x08password\x18\x02 \x01(\tR\x08password\x12\x1a\n\x08Keyspace\x18\x03 \x01(\tR\x08Keyspace\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a#\n\x03Res\x12\x1c\n\tsessionId\x18\x01 \x01(\tR\tsessionId\x1a\xb8\x01\n\x05\x43lose\x1a\xa7\x01\n\x03Req\x12\x45\n\x08metadata\x18\xe8\x07 \x03(\x0b\x32(.session.Session.Close.Req.MetadataEntryR\x08metadata\x12\x1c\n\tsessionId\x18\x01 \x01(\tR\tsessionId\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x05\n\x03Res\"\xfd \n\x0bTransaction\x1a\xcc\t\n\x03Req\x12\x43\n\x08metadata\x18\xe8\x07 \x03(\x0b\x32&.session.Transaction.Req.MetadataEntryR\x08metadata\x12:\n\x08open_req\x18\x01 \x01(\x0b\x32\x1d.session.Transaction.Open.ReqH\x00R\x07openReq\x12@\n\ncommit_req\x18\x02 \x01(\x0b\x32\x1f.session.Transaction.Commit.ReqH\x00R\tcommitReq\x12=\n\tquery_req\x18\x03 \x01(\x0b\x32\x1e.session.Transaction.Query.ReqH\x00R\x08queryReq\x12@\n\x0biterate_req\x18\x04 \x01(\x0b\x32\x1d.session.Transaction.Iter.ReqH\x00R\niterateReq\x12^\n\x14getSchemaConcept_req\x18\x05 \x01(\x0b\x32).session.Transaction.GetSchemaConcept.ReqH\x00R\x13getSchemaConceptReq\x12L\n\x0egetConcept_req\x18\x06 \x01(\x0b\x32#.session.Transaction.GetConcept.ReqH\x00R\rgetConceptReq\x12U\n\x11getAttributes_req\x18\x07 \x01(\x0b\x32&.session.Transaction.GetAttributes.ReqH\x00R\x10getAttributesReq\x12U\n\x11putEntityType_req\x18\x08 \x01(\x0b\x32&.session.Transaction.PutEntityType.ReqH\x00R\x10putEntityTypeReq\x12^\n\x14putAttributeType_req\x18\t \x01(\x0b\x32).session.Transaction.PutAttributeType.ReqH\x00R\x13putAttributeTypeReq\x12[\n\x13putRelationType_req\x18\n \x01(\x0b\x32(.session.Transaction.PutRelationType.ReqH\x00R\x12putRelationTypeReq\x12\x43\n\x0bputRole_req\x18\x0b \x01(\x0b\x32 .session.Transaction.PutRole.ReqH\x00R\nputRoleReq\x12\x43\n\x0bputRule_req\x18\x0c \x01(\x0b\x32 .session.Transaction.PutRule.ReqH\x00R\nputRuleReq\x12U\n\x11\x63onceptMethod_req\x18\r \x01(\x0b\x32&.session.Transaction.ConceptMethod.ReqH\x00R\x10\x63onceptMethodReq\x12\x43\n\x0f\x65xplanation_req\x18\x0e \x01(\x0b\x32\x18.session.Explanation.ReqH\x00R\x0e\x65xplanationReq\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x05\n\x03req\x1a\xd0\x08\n\x03Res\x12:\n\x08open_res\x18\x01 \x01(\x0b\x32\x1d.session.Transaction.Open.ResH\x00R\x07openRes\x12@\n\ncommit_res\x18\x02 \x01(\x0b\x32\x1f.session.Transaction.Commit.ResH\x00R\tcommitRes\x12@\n\nquery_iter\x18\x03 \x01(\x0b\x32\x1f.session.Transaction.Query.IterH\x00R\tqueryIter\x12@\n\x0biterate_res\x18\x04 \x01(\x0b\x32\x1d.session.Transaction.Iter.ResH\x00R\niterateRes\x12^\n\x14getSchemaConcept_res\x18\x05 \x01(\x0b\x32).session.Transaction.GetSchemaConcept.ResH\x00R\x13getSchemaConceptRes\x12L\n\x0egetConcept_res\x18\x06 \x01(\x0b\x32#.session.Transaction.GetConcept.ResH\x00R\rgetConceptRes\x12X\n\x12getAttributes_iter\x18\x07 \x01(\x0b\x32\'.session.Transaction.GetAttributes.IterH\x00R\x11getAttributesIter\x12U\n\x11putEntityType_res\x18\x08 \x01(\x0b\x32&.session.Transaction.PutEntityType.ResH\x00R\x10putEntityTypeRes\x12^\n\x14putAttributeType_res\x18\t \x01(\x0b\x32).session.Transaction.PutAttributeType.ResH\x00R\x13putAttributeTypeRes\x12[\n\x13putRelationType_res\x18\n \x01(\x0b\x32(.session.Transaction.PutRelationType.ResH\x00R\x12putRelationTypeRes\x12\x43\n\x0bputRole_res\x18\x0b \x01(\x0b\x32 .session.Transaction.PutRole.ResH\x00R\nputRoleRes\x12\x43\n\x0bputRule_res\x18\x0c \x01(\x0b\x32 .session.Transaction.PutRule.ResH\x00R\nputRuleRes\x12U\n\x11\x63onceptMethod_res\x18\r \x01(\x0b\x32&.session.Transaction.ConceptMethod.ResH\x00R\x10\x63onceptMethodRes\x12\x43\n\x0f\x65xplanation_res\x18\x0e \x01(\x0b\x32\x18.session.Explanation.ResH\x00R\x0e\x65xplanationResB\x05\n\x03res\x1a\xc6\x02\n\x04Iter\x1a\x15\n\x03Req\x12\x0e\n\x02id\x18\x01 \x01(\x05R\x02id\x1a\xa6\x02\n\x03Res\x12\x14\n\x04\x64one\x18\x01 \x01(\x08H\x00R\x04\x64one\x12K\n\x0equery_iter_res\x18\x02 \x01(\x0b\x32#.session.Transaction.Query.Iter.ResH\x00R\x0cqueryIterRes\x12\x63\n\x16getAttributes_iter_res\x18\x03 \x01(\x0b\x32+.session.Transaction.GetAttributes.Iter.ResH\x00R\x14getAttributesIterRes\x12P\n\x16\x63onceptMethod_iter_res\x18\x04 \x01(\x0b\x32\x18.session.Method.Iter.ResH\x00R\x14\x63onceptMethodIterResB\x05\n\x03res\x1a\x61\n\x04Open\x1aR\n\x03Req\x12\x1c\n\tsessionId\x18\x01 \x01(\tR\tsessionId\x12-\n\x04type\x18\x02 \x01(\x0e\x32\x19.session.Transaction.TypeR\x04type\x1a\x05\n\x03Res\x1a\x16\n\x06\x43ommit\x1a\x05\n\x03Req\x1a\x05\n\x03Res\x1a\xc2\x01\n\x05Query\x1aS\n\x03Req\x12\x14\n\x05query\x18\x01 \x01(\tR\x05query\x12\x36\n\x05infer\x18\x02 \x01(\x0e\x32 .session.Transaction.Query.INFERR\x05infer\x1a\x46\n\x04Iter\x12\x0e\n\x02id\x18\x01 \x01(\x05R\x02id\x1a.\n\x03Res\x12\'\n\x06\x61nswer\x18\x01 \x01(\x0b\x32\x0f.session.AnswerR\x06\x61nswer\"\x1c\n\x05INFER\x12\x08\n\x04TRUE\x10\x00\x12\t\n\x05\x46\x41LSE\x10\x01\x1a\x9c\x01\n\x10GetSchemaConcept\x1a\x1b\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x1ak\n\x03Res\x12\x38\n\rschemaConcept\x18\x01 \x01(\x0b\x32\x10.session.ConceptH\x00R\rschemaConcept\x12#\n\x04null\x18\x02 \x01(\x0b\x32\r.session.NullH\x00R\x04nullB\x05\n\x03res\x1a\x84\x01\n\nGetConcept\x1a\x15\n\x03Req\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x1a_\n\x03Res\x12,\n\x07\x63oncept\x18\x01 \x01(\x0b\x32\x10.session.ConceptH\x00R\x07\x63oncept\x12#\n\x04null\x18\x02 \x01(\x0b\x32\r.session.NullH\x00R\x04nullB\x05\n\x03res\x1a\x91\x01\n\rGetAttributes\x1a\x31\n\x03Req\x12*\n\x05value\x18\x01 \x01(\x0b\x32\x14.session.ValueObjectR\x05value\x1aM\n\x04Iter\x12\x0e\n\x02id\x18\x01 \x01(\x05R\x02id\x1a\x35\n\x03Res\x12.\n\tattribute\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\tattribute\x1a\x65\n\rPutEntityType\x1a\x1b\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x1a\x37\n\x03Res\x12\x30\n\nentityType\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\nentityType\x1a\xac\x01\n\x10PutAttributeType\x1aY\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x12<\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32 .session.AttributeType.DATA_TYPER\x08\x64\x61taType\x1a=\n\x03Res\x12\x36\n\rattributeType\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\rattributeType\x1ak\n\x0fPutRelationType\x1a\x1b\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x1a;\n\x03Res\x12\x34\n\x0crelationType\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\x0crelationType\x1aS\n\x07PutRole\x1a\x1b\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x1a+\n\x03Res\x12$\n\x04role\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\x04role\x1a{\n\x07PutRule\x1a\x43\n\x03Req\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x12\x12\n\x04when\x18\x02 \x01(\tR\x04when\x12\x12\n\x04then\x18\x03 \x01(\tR\x04then\x1a+\n\x03Res\x12$\n\x04rule\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\x04rule\x1a\x8b\x01\n\rConceptMethod\x1a\x42\n\x03Req\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12+\n\x06method\x18\x02 \x01(\x0b\x32\x13.session.Method.ReqR\x06method\x1a\x36\n\x03Res\x12/\n\x08response\x18\x01 \x01(\x0b\x32\x13.session.Method.ResR\x08response\"&\n\x04Type\x12\x08\n\x04READ\x10\x00\x12\t\n\x05WRITE\x10\x01\x12\t\n\x05\x42\x41TCH\x10\x02\x32\xd6\x01\n\x0eSessionService\x12<\n\x04open\x12\x19.session.Session.Open.Req\x1a\x19.session.Session.Open.Res\x12?\n\x05\x63lose\x12\x1a.session.Session.Close.Req\x1a\x1a.session.Session.Close.Res\x12\x45\n\x0btransaction\x12\x18.session.Transaction.Req\x1a\x18.session.Transaction.Res(\x01\x30\x01\x42&\n\x16grakn.protocol.sessionB\x0cSessionProtob\x06proto3')
  ,
  dependencies=[session_dot_Concept__pb2.DESCRIPTOR,session_dot_Answer__pb2.DESCRIPTOR,])



_TRANSACTION_QUERY_INFER = _descriptor.EnumDescriptor(
  name='INFER',
  full_name='session.Transaction.Query.INFER',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TRUE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FALSE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3520,
  serialized_end=3548,
)
_sym_db.RegisterEnumDescriptor(_TRANSACTION_QUERY_INFER)

_TRANSACTION_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='session.Transaction.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='READ', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WRITE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BATCH', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=4731,
  serialized_end=4769,
)
_sym_db.RegisterEnumDescriptor(_TRANSACTION_TYPE)


_SESSION_OPEN_REQ_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='session.Session.Open.Req.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='session.Session.Open.Req.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Session.Open.Req.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=321,
)

_SESSION_OPEN_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Session.Open.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='session.Session.Open.Req.metadata', index=0,
      number=1000, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='metadata', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='username', full_name='session.Session.Open.Req.username', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='username', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='password', full_name='session.Session.Open.Req.password', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='password', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Keyspace', full_name='session.Session.Open.Req.Keyspace', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='Keyspace', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SESSION_OPEN_REQ_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=321,
)

_SESSION_OPEN_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Session.Open.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sessionId', full_name='session.Session.Open.Res.sessionId', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sessionId', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=358,
)

_SESSION_OPEN = _descriptor.Descriptor(
  name='Open',
  full_name='session.Session.Open',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_SESSION_OPEN_REQ, _SESSION_OPEN_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=92,
  serialized_end=358,
)

_SESSION_CLOSE_REQ_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='session.Session.Close.Req.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='session.Session.Close.Req.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Session.Close.Req.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=321,
)

_SESSION_CLOSE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Session.Close.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='session.Session.Close.Req.metadata', index=0,
      number=1000, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='metadata', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sessionId', full_name='session.Session.Close.Req.sessionId', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sessionId', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SESSION_CLOSE_REQ_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=371,
  serialized_end=538,
)

_SESSION_CLOSE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Session.Close.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=328,
)

_SESSION_CLOSE = _descriptor.Descriptor(
  name='Close',
  full_name='session.Session.Close',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_SESSION_CLOSE_REQ, _SESSION_CLOSE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=361,
  serialized_end=545,
)

_SESSION = _descriptor.Descriptor(
  name='Session',
  full_name='session.Session',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_SESSION_OPEN, _SESSION_CLOSE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=545,
)


_TRANSACTION_REQ_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='session.Transaction.Req.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='session.Transaction.Req.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Transaction.Req.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=321,
)

_TRANSACTION_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='session.Transaction.Req.metadata', index=0,
      number=1000, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='metadata', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='open_req', full_name='session.Transaction.Req.open_req', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='openReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit_req', full_name='session.Transaction.Req.commit_req', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='commitReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_req', full_name='session.Transaction.Req.query_req', index=3,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='queryReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='iterate_req', full_name='session.Transaction.Req.iterate_req', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='iterateReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getSchemaConcept_req', full_name='session.Transaction.Req.getSchemaConcept_req', index=5,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getSchemaConceptReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getConcept_req', full_name='session.Transaction.Req.getConcept_req', index=6,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getConceptReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getAttributes_req', full_name='session.Transaction.Req.getAttributes_req', index=7,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getAttributesReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putEntityType_req', full_name='session.Transaction.Req.putEntityType_req', index=8,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putEntityTypeReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putAttributeType_req', full_name='session.Transaction.Req.putAttributeType_req', index=9,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putAttributeTypeReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRelationType_req', full_name='session.Transaction.Req.putRelationType_req', index=10,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRelationTypeReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRole_req', full_name='session.Transaction.Req.putRole_req', index=11,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRoleReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRule_req', full_name='session.Transaction.Req.putRule_req', index=12,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRuleReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptMethod_req', full_name='session.Transaction.Req.conceptMethod_req', index=13,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptMethodReq', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explanation_req', full_name='session.Transaction.Req.explanation_req', index=14,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='explanationReq', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_REQ_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='req', full_name='session.Transaction.Req.req',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=564,
  serialized_end=1792,
)

_TRANSACTION_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='open_res', full_name='session.Transaction.Res.open_res', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='openRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit_res', full_name='session.Transaction.Res.commit_res', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='commitRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_iter', full_name='session.Transaction.Res.query_iter', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='queryIter', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='iterate_res', full_name='session.Transaction.Res.iterate_res', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='iterateRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getSchemaConcept_res', full_name='session.Transaction.Res.getSchemaConcept_res', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getSchemaConceptRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getConcept_res', full_name='session.Transaction.Res.getConcept_res', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getConceptRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getAttributes_iter', full_name='session.Transaction.Res.getAttributes_iter', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getAttributesIter', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putEntityType_res', full_name='session.Transaction.Res.putEntityType_res', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putEntityTypeRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putAttributeType_res', full_name='session.Transaction.Res.putAttributeType_res', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putAttributeTypeRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRelationType_res', full_name='session.Transaction.Res.putRelationType_res', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRelationTypeRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRole_res', full_name='session.Transaction.Res.putRole_res', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRoleRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='putRule_res', full_name='session.Transaction.Res.putRule_res', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='putRuleRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptMethod_res', full_name='session.Transaction.Res.conceptMethod_res', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptMethodRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explanation_res', full_name='session.Transaction.Res.explanation_res', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='explanationRes', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='res', full_name='session.Transaction.Res.res',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1795,
  serialized_end=2899,
)

_TRANSACTION_ITER_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.Iter.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='session.Transaction.Iter.Req.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2910,
  serialized_end=2931,
)

_TRANSACTION_ITER_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.Iter.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='done', full_name='session.Transaction.Iter.Res.done', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='done', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_iter_res', full_name='session.Transaction.Iter.Res.query_iter_res', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='queryIterRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='getAttributes_iter_res', full_name='session.Transaction.Iter.Res.getAttributes_iter_res', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='getAttributesIterRes', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptMethod_iter_res', full_name='session.Transaction.Iter.Res.conceptMethod_iter_res', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptMethodIterRes', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='res', full_name='session.Transaction.Iter.Res.res',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=2934,
  serialized_end=3228,
)

_TRANSACTION_ITER = _descriptor.Descriptor(
  name='Iter',
  full_name='session.Transaction.Iter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_ITER_REQ, _TRANSACTION_ITER_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2902,
  serialized_end=3228,
)

_TRANSACTION_OPEN_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.Open.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sessionId', full_name='session.Transaction.Open.Req.sessionId', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sessionId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='session.Transaction.Open.Req.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='type', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3238,
  serialized_end=3320,
)

_TRANSACTION_OPEN_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.Open.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=328,
)

_TRANSACTION_OPEN = _descriptor.Descriptor(
  name='Open',
  full_name='session.Transaction.Open',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_OPEN_REQ, _TRANSACTION_OPEN_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3230,
  serialized_end=3327,
)

_TRANSACTION_COMMIT_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.Commit.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=106,
)

_TRANSACTION_COMMIT_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.Commit.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=328,
)

_TRANSACTION_COMMIT = _descriptor.Descriptor(
  name='Commit',
  full_name='session.Transaction.Commit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_COMMIT_REQ, _TRANSACTION_COMMIT_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3329,
  serialized_end=3351,
)

_TRANSACTION_QUERY_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.Query.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='session.Transaction.Query.Req.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='query', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='infer', full_name='session.Transaction.Query.Req.infer', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='infer', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3363,
  serialized_end=3446,
)

_TRANSACTION_QUERY_ITER_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.Query.Iter.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='answer', full_name='session.Transaction.Query.Iter.Res.answer', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='answer', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3472,
  serialized_end=3518,
)

_TRANSACTION_QUERY_ITER = _descriptor.Descriptor(
  name='Iter',
  full_name='session.Transaction.Query.Iter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='session.Transaction.Query.Iter.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_QUERY_ITER_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3448,
  serialized_end=3518,
)

_TRANSACTION_QUERY = _descriptor.Descriptor(
  name='Query',
  full_name='session.Transaction.Query',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_QUERY_REQ, _TRANSACTION_QUERY_ITER, ],
  enum_types=[
    _TRANSACTION_QUERY_INFER,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3354,
  serialized_end=3548,
)

_TRANSACTION_GETSCHEMACONCEPT_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.GetSchemaConcept.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.GetSchemaConcept.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3571,
  serialized_end=3598,
)

_TRANSACTION_GETSCHEMACONCEPT_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.GetSchemaConcept.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='schemaConcept', full_name='session.Transaction.GetSchemaConcept.Res.schemaConcept', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='schemaConcept', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='null', full_name='session.Transaction.GetSchemaConcept.Res.null', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='null', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='res', full_name='session.Transaction.GetSchemaConcept.Res.res',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=3600,
  serialized_end=3707,
)

_TRANSACTION_GETSCHEMACONCEPT = _descriptor.Descriptor(
  name='GetSchemaConcept',
  full_name='session.Transaction.GetSchemaConcept',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_GETSCHEMACONCEPT_REQ, _TRANSACTION_GETSCHEMACONCEPT_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3551,
  serialized_end=3707,
)

_TRANSACTION_GETCONCEPT_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.GetConcept.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='session.Transaction.GetConcept.Req.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3724,
  serialized_end=3745,
)

_TRANSACTION_GETCONCEPT_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.GetConcept.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='concept', full_name='session.Transaction.GetConcept.Res.concept', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='concept', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='null', full_name='session.Transaction.GetConcept.Res.null', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='null', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='res', full_name='session.Transaction.GetConcept.Res.res',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=3747,
  serialized_end=3842,
)

_TRANSACTION_GETCONCEPT = _descriptor.Descriptor(
  name='GetConcept',
  full_name='session.Transaction.GetConcept',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_GETCONCEPT_REQ, _TRANSACTION_GETCONCEPT_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3710,
  serialized_end=3842,
)

_TRANSACTION_GETATTRIBUTES_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.GetAttributes.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Transaction.GetAttributes.Req.value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3862,
  serialized_end=3911,
)

_TRANSACTION_GETATTRIBUTES_ITER_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.GetAttributes.Iter.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='attribute', full_name='session.Transaction.GetAttributes.Iter.Res.attribute', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='attribute', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3937,
  serialized_end=3990,
)

_TRANSACTION_GETATTRIBUTES_ITER = _descriptor.Descriptor(
  name='Iter',
  full_name='session.Transaction.GetAttributes.Iter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='session.Transaction.GetAttributes.Iter.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_GETATTRIBUTES_ITER_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3913,
  serialized_end=3990,
)

_TRANSACTION_GETATTRIBUTES = _descriptor.Descriptor(
  name='GetAttributes',
  full_name='session.Transaction.GetAttributes',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_GETATTRIBUTES_REQ, _TRANSACTION_GETATTRIBUTES_ITER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3845,
  serialized_end=3990,
)

_TRANSACTION_PUTENTITYTYPE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.PutEntityType.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.PutEntityType.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3571,
  serialized_end=3598,
)

_TRANSACTION_PUTENTITYTYPE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.PutEntityType.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entityType', full_name='session.Transaction.PutEntityType.Res.entityType', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='entityType', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4038,
  serialized_end=4093,
)

_TRANSACTION_PUTENTITYTYPE = _descriptor.Descriptor(
  name='PutEntityType',
  full_name='session.Transaction.PutEntityType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PUTENTITYTYPE_REQ, _TRANSACTION_PUTENTITYTYPE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3992,
  serialized_end=4093,
)

_TRANSACTION_PUTATTRIBUTETYPE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.PutAttributeType.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.PutAttributeType.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dataType', full_name='session.Transaction.PutAttributeType.Req.dataType', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='dataType', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4116,
  serialized_end=4205,
)

_TRANSACTION_PUTATTRIBUTETYPE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.PutAttributeType.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='attributeType', full_name='session.Transaction.PutAttributeType.Res.attributeType', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='attributeType', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4207,
  serialized_end=4268,
)

_TRANSACTION_PUTATTRIBUTETYPE = _descriptor.Descriptor(
  name='PutAttributeType',
  full_name='session.Transaction.PutAttributeType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PUTATTRIBUTETYPE_REQ, _TRANSACTION_PUTATTRIBUTETYPE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4096,
  serialized_end=4268,
)

_TRANSACTION_PUTRELATIONTYPE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.PutRelationType.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.PutRelationType.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3571,
  serialized_end=3598,
)

_TRANSACTION_PUTRELATIONTYPE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.PutRelationType.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='relationType', full_name='session.Transaction.PutRelationType.Res.relationType', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='relationType', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4318,
  serialized_end=4377,
)

_TRANSACTION_PUTRELATIONTYPE = _descriptor.Descriptor(
  name='PutRelationType',
  full_name='session.Transaction.PutRelationType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PUTRELATIONTYPE_REQ, _TRANSACTION_PUTRELATIONTYPE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4270,
  serialized_end=4377,
)

_TRANSACTION_PUTROLE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.PutRole.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.PutRole.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3571,
  serialized_end=3598,
)

_TRANSACTION_PUTROLE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.PutRole.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='role', full_name='session.Transaction.PutRole.Res.role', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='role', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4419,
  serialized_end=4462,
)

_TRANSACTION_PUTROLE = _descriptor.Descriptor(
  name='PutRole',
  full_name='session.Transaction.PutRole',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PUTROLE_REQ, _TRANSACTION_PUTROLE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4379,
  serialized_end=4462,
)

_TRANSACTION_PUTRULE_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.PutRule.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='session.Transaction.PutRule.Req.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='label', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='when', full_name='session.Transaction.PutRule.Req.when', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='when', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='then', full_name='session.Transaction.PutRule.Req.then', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='then', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4475,
  serialized_end=4542,
)

_TRANSACTION_PUTRULE_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.PutRule.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rule', full_name='session.Transaction.PutRule.Res.rule', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='rule', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4544,
  serialized_end=4587,
)

_TRANSACTION_PUTRULE = _descriptor.Descriptor(
  name='PutRule',
  full_name='session.Transaction.PutRule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PUTRULE_REQ, _TRANSACTION_PUTRULE_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4464,
  serialized_end=4587,
)

_TRANSACTION_CONCEPTMETHOD_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Transaction.ConceptMethod.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='session.Transaction.ConceptMethod.Req.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='session.Transaction.ConceptMethod.Req.method', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='method', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4607,
  serialized_end=4673,
)

_TRANSACTION_CONCEPTMETHOD_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Transaction.ConceptMethod.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response', full_name='session.Transaction.ConceptMethod.Res.response', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='response', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4675,
  serialized_end=4729,
)

_TRANSACTION_CONCEPTMETHOD = _descriptor.Descriptor(
  name='ConceptMethod',
  full_name='session.Transaction.ConceptMethod',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_CONCEPTMETHOD_REQ, _TRANSACTION_CONCEPTMETHOD_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4590,
  serialized_end=4729,
)

_TRANSACTION = _descriptor.Descriptor(
  name='Transaction',
  full_name='session.Transaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_REQ, _TRANSACTION_RES, _TRANSACTION_ITER, _TRANSACTION_OPEN, _TRANSACTION_COMMIT, _TRANSACTION_QUERY, _TRANSACTION_GETSCHEMACONCEPT, _TRANSACTION_GETCONCEPT, _TRANSACTION_GETATTRIBUTES, _TRANSACTION_PUTENTITYTYPE, _TRANSACTION_PUTATTRIBUTETYPE, _TRANSACTION_PUTRELATIONTYPE, _TRANSACTION_PUTROLE, _TRANSACTION_PUTRULE, _TRANSACTION_CONCEPTMETHOD, ],
  enum_types=[
    _TRANSACTION_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=548,
  serialized_end=4769,
)

_SESSION_OPEN_REQ_METADATAENTRY.containing_type = _SESSION_OPEN_REQ
_SESSION_OPEN_REQ.fields_by_name['metadata'].message_type = _SESSION_OPEN_REQ_METADATAENTRY
_SESSION_OPEN_REQ.containing_type = _SESSION_OPEN
_SESSION_OPEN_RES.containing_type = _SESSION_OPEN
_SESSION_OPEN.containing_type = _SESSION
_SESSION_CLOSE_REQ_METADATAENTRY.containing_type = _SESSION_CLOSE_REQ
_SESSION_CLOSE_REQ.fields_by_name['metadata'].message_type = _SESSION_CLOSE_REQ_METADATAENTRY
_SESSION_CLOSE_REQ.containing_type = _SESSION_CLOSE
_SESSION_CLOSE_RES.containing_type = _SESSION_CLOSE
_SESSION_CLOSE.containing_type = _SESSION
_TRANSACTION_REQ_METADATAENTRY.containing_type = _TRANSACTION_REQ
_TRANSACTION_REQ.fields_by_name['metadata'].message_type = _TRANSACTION_REQ_METADATAENTRY
_TRANSACTION_REQ.fields_by_name['open_req'].message_type = _TRANSACTION_OPEN_REQ
_TRANSACTION_REQ.fields_by_name['commit_req'].message_type = _TRANSACTION_COMMIT_REQ
_TRANSACTION_REQ.fields_by_name['query_req'].message_type = _TRANSACTION_QUERY_REQ
_TRANSACTION_REQ.fields_by_name['iterate_req'].message_type = _TRANSACTION_ITER_REQ
_TRANSACTION_REQ.fields_by_name['getSchemaConcept_req'].message_type = _TRANSACTION_GETSCHEMACONCEPT_REQ
_TRANSACTION_REQ.fields_by_name['getConcept_req'].message_type = _TRANSACTION_GETCONCEPT_REQ
_TRANSACTION_REQ.fields_by_name['getAttributes_req'].message_type = _TRANSACTION_GETATTRIBUTES_REQ
_TRANSACTION_REQ.fields_by_name['putEntityType_req'].message_type = _TRANSACTION_PUTENTITYTYPE_REQ
_TRANSACTION_REQ.fields_by_name['putAttributeType_req'].message_type = _TRANSACTION_PUTATTRIBUTETYPE_REQ
_TRANSACTION_REQ.fields_by_name['putRelationType_req'].message_type = _TRANSACTION_PUTRELATIONTYPE_REQ
_TRANSACTION_REQ.fields_by_name['putRole_req'].message_type = _TRANSACTION_PUTROLE_REQ
_TRANSACTION_REQ.fields_by_name['putRule_req'].message_type = _TRANSACTION_PUTRULE_REQ
_TRANSACTION_REQ.fields_by_name['conceptMethod_req'].message_type = _TRANSACTION_CONCEPTMETHOD_REQ
_TRANSACTION_REQ.fields_by_name['explanation_req'].message_type = session_dot_Answer__pb2._EXPLANATION_REQ
_TRANSACTION_REQ.containing_type = _TRANSACTION
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['open_req'])
_TRANSACTION_REQ.fields_by_name['open_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['commit_req'])
_TRANSACTION_REQ.fields_by_name['commit_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['query_req'])
_TRANSACTION_REQ.fields_by_name['query_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['iterate_req'])
_TRANSACTION_REQ.fields_by_name['iterate_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['getSchemaConcept_req'])
_TRANSACTION_REQ.fields_by_name['getSchemaConcept_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['getConcept_req'])
_TRANSACTION_REQ.fields_by_name['getConcept_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['getAttributes_req'])
_TRANSACTION_REQ.fields_by_name['getAttributes_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['putEntityType_req'])
_TRANSACTION_REQ.fields_by_name['putEntityType_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['putAttributeType_req'])
_TRANSACTION_REQ.fields_by_name['putAttributeType_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['putRelationType_req'])
_TRANSACTION_REQ.fields_by_name['putRelationType_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['putRole_req'])
_TRANSACTION_REQ.fields_by_name['putRole_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['putRule_req'])
_TRANSACTION_REQ.fields_by_name['putRule_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['conceptMethod_req'])
_TRANSACTION_REQ.fields_by_name['conceptMethod_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_REQ.oneofs_by_name['req'].fields.append(
  _TRANSACTION_REQ.fields_by_name['explanation_req'])
_TRANSACTION_REQ.fields_by_name['explanation_req'].containing_oneof = _TRANSACTION_REQ.oneofs_by_name['req']
_TRANSACTION_RES.fields_by_name['open_res'].message_type = _TRANSACTION_OPEN_RES
_TRANSACTION_RES.fields_by_name['commit_res'].message_type = _TRANSACTION_COMMIT_RES
_TRANSACTION_RES.fields_by_name['query_iter'].message_type = _TRANSACTION_QUERY_ITER
_TRANSACTION_RES.fields_by_name['iterate_res'].message_type = _TRANSACTION_ITER_RES
_TRANSACTION_RES.fields_by_name['getSchemaConcept_res'].message_type = _TRANSACTION_GETSCHEMACONCEPT_RES
_TRANSACTION_RES.fields_by_name['getConcept_res'].message_type = _TRANSACTION_GETCONCEPT_RES
_TRANSACTION_RES.fields_by_name['getAttributes_iter'].message_type = _TRANSACTION_GETATTRIBUTES_ITER
_TRANSACTION_RES.fields_by_name['putEntityType_res'].message_type = _TRANSACTION_PUTENTITYTYPE_RES
_TRANSACTION_RES.fields_by_name['putAttributeType_res'].message_type = _TRANSACTION_PUTATTRIBUTETYPE_RES
_TRANSACTION_RES.fields_by_name['putRelationType_res'].message_type = _TRANSACTION_PUTRELATIONTYPE_RES
_TRANSACTION_RES.fields_by_name['putRole_res'].message_type = _TRANSACTION_PUTROLE_RES
_TRANSACTION_RES.fields_by_name['putRule_res'].message_type = _TRANSACTION_PUTRULE_RES
_TRANSACTION_RES.fields_by_name['conceptMethod_res'].message_type = _TRANSACTION_CONCEPTMETHOD_RES
_TRANSACTION_RES.fields_by_name['explanation_res'].message_type = session_dot_Answer__pb2._EXPLANATION_RES
_TRANSACTION_RES.containing_type = _TRANSACTION
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['open_res'])
_TRANSACTION_RES.fields_by_name['open_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['commit_res'])
_TRANSACTION_RES.fields_by_name['commit_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['query_iter'])
_TRANSACTION_RES.fields_by_name['query_iter'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['iterate_res'])
_TRANSACTION_RES.fields_by_name['iterate_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['getSchemaConcept_res'])
_TRANSACTION_RES.fields_by_name['getSchemaConcept_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['getConcept_res'])
_TRANSACTION_RES.fields_by_name['getConcept_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['getAttributes_iter'])
_TRANSACTION_RES.fields_by_name['getAttributes_iter'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['putEntityType_res'])
_TRANSACTION_RES.fields_by_name['putEntityType_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['putAttributeType_res'])
_TRANSACTION_RES.fields_by_name['putAttributeType_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['putRelationType_res'])
_TRANSACTION_RES.fields_by_name['putRelationType_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['putRole_res'])
_TRANSACTION_RES.fields_by_name['putRole_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['putRule_res'])
_TRANSACTION_RES.fields_by_name['putRule_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['conceptMethod_res'])
_TRANSACTION_RES.fields_by_name['conceptMethod_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_RES.fields_by_name['explanation_res'])
_TRANSACTION_RES.fields_by_name['explanation_res'].containing_oneof = _TRANSACTION_RES.oneofs_by_name['res']
_TRANSACTION_ITER_REQ.containing_type = _TRANSACTION_ITER
_TRANSACTION_ITER_RES.fields_by_name['query_iter_res'].message_type = _TRANSACTION_QUERY_ITER_RES
_TRANSACTION_ITER_RES.fields_by_name['getAttributes_iter_res'].message_type = _TRANSACTION_GETATTRIBUTES_ITER_RES
_TRANSACTION_ITER_RES.fields_by_name['conceptMethod_iter_res'].message_type = session_dot_Concept__pb2._METHOD_ITER_RES
_TRANSACTION_ITER_RES.containing_type = _TRANSACTION_ITER
_TRANSACTION_ITER_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_ITER_RES.fields_by_name['done'])
_TRANSACTION_ITER_RES.fields_by_name['done'].containing_oneof = _TRANSACTION_ITER_RES.oneofs_by_name['res']
_TRANSACTION_ITER_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_ITER_RES.fields_by_name['query_iter_res'])
_TRANSACTION_ITER_RES.fields_by_name['query_iter_res'].containing_oneof = _TRANSACTION_ITER_RES.oneofs_by_name['res']
_TRANSACTION_ITER_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_ITER_RES.fields_by_name['getAttributes_iter_res'])
_TRANSACTION_ITER_RES.fields_by_name['getAttributes_iter_res'].containing_oneof = _TRANSACTION_ITER_RES.oneofs_by_name['res']
_TRANSACTION_ITER_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_ITER_RES.fields_by_name['conceptMethod_iter_res'])
_TRANSACTION_ITER_RES.fields_by_name['conceptMethod_iter_res'].containing_oneof = _TRANSACTION_ITER_RES.oneofs_by_name['res']
_TRANSACTION_ITER.containing_type = _TRANSACTION
_TRANSACTION_OPEN_REQ.fields_by_name['type'].enum_type = _TRANSACTION_TYPE
_TRANSACTION_OPEN_REQ.containing_type = _TRANSACTION_OPEN
_TRANSACTION_OPEN_RES.containing_type = _TRANSACTION_OPEN
_TRANSACTION_OPEN.containing_type = _TRANSACTION
_TRANSACTION_COMMIT_REQ.containing_type = _TRANSACTION_COMMIT
_TRANSACTION_COMMIT_RES.containing_type = _TRANSACTION_COMMIT
_TRANSACTION_COMMIT.containing_type = _TRANSACTION
_TRANSACTION_QUERY_REQ.fields_by_name['infer'].enum_type = _TRANSACTION_QUERY_INFER
_TRANSACTION_QUERY_REQ.containing_type = _TRANSACTION_QUERY
_TRANSACTION_QUERY_ITER_RES.fields_by_name['answer'].message_type = session_dot_Answer__pb2._ANSWER
_TRANSACTION_QUERY_ITER_RES.containing_type = _TRANSACTION_QUERY_ITER
_TRANSACTION_QUERY_ITER.containing_type = _TRANSACTION_QUERY
_TRANSACTION_QUERY.containing_type = _TRANSACTION
_TRANSACTION_QUERY_INFER.containing_type = _TRANSACTION_QUERY
_TRANSACTION_GETSCHEMACONCEPT_REQ.containing_type = _TRANSACTION_GETSCHEMACONCEPT
_TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['schemaConcept'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['null'].message_type = session_dot_Concept__pb2._NULL
_TRANSACTION_GETSCHEMACONCEPT_RES.containing_type = _TRANSACTION_GETSCHEMACONCEPT
_TRANSACTION_GETSCHEMACONCEPT_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['schemaConcept'])
_TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['schemaConcept'].containing_oneof = _TRANSACTION_GETSCHEMACONCEPT_RES.oneofs_by_name['res']
_TRANSACTION_GETSCHEMACONCEPT_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['null'])
_TRANSACTION_GETSCHEMACONCEPT_RES.fields_by_name['null'].containing_oneof = _TRANSACTION_GETSCHEMACONCEPT_RES.oneofs_by_name['res']
_TRANSACTION_GETSCHEMACONCEPT.containing_type = _TRANSACTION
_TRANSACTION_GETCONCEPT_REQ.containing_type = _TRANSACTION_GETCONCEPT
_TRANSACTION_GETCONCEPT_RES.fields_by_name['concept'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_GETCONCEPT_RES.fields_by_name['null'].message_type = session_dot_Concept__pb2._NULL
_TRANSACTION_GETCONCEPT_RES.containing_type = _TRANSACTION_GETCONCEPT
_TRANSACTION_GETCONCEPT_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_GETCONCEPT_RES.fields_by_name['concept'])
_TRANSACTION_GETCONCEPT_RES.fields_by_name['concept'].containing_oneof = _TRANSACTION_GETCONCEPT_RES.oneofs_by_name['res']
_TRANSACTION_GETCONCEPT_RES.oneofs_by_name['res'].fields.append(
  _TRANSACTION_GETCONCEPT_RES.fields_by_name['null'])
_TRANSACTION_GETCONCEPT_RES.fields_by_name['null'].containing_oneof = _TRANSACTION_GETCONCEPT_RES.oneofs_by_name['res']
_TRANSACTION_GETCONCEPT.containing_type = _TRANSACTION
_TRANSACTION_GETATTRIBUTES_REQ.fields_by_name['value'].message_type = session_dot_Concept__pb2._VALUEOBJECT
_TRANSACTION_GETATTRIBUTES_REQ.containing_type = _TRANSACTION_GETATTRIBUTES
_TRANSACTION_GETATTRIBUTES_ITER_RES.fields_by_name['attribute'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_GETATTRIBUTES_ITER_RES.containing_type = _TRANSACTION_GETATTRIBUTES_ITER
_TRANSACTION_GETATTRIBUTES_ITER.containing_type = _TRANSACTION_GETATTRIBUTES
_TRANSACTION_GETATTRIBUTES.containing_type = _TRANSACTION
_TRANSACTION_PUTENTITYTYPE_REQ.containing_type = _TRANSACTION_PUTENTITYTYPE
_TRANSACTION_PUTENTITYTYPE_RES.fields_by_name['entityType'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_PUTENTITYTYPE_RES.containing_type = _TRANSACTION_PUTENTITYTYPE
_TRANSACTION_PUTENTITYTYPE.containing_type = _TRANSACTION
_TRANSACTION_PUTATTRIBUTETYPE_REQ.fields_by_name['dataType'].enum_type = session_dot_Concept__pb2._ATTRIBUTETYPE_DATA_TYPE
_TRANSACTION_PUTATTRIBUTETYPE_REQ.containing_type = _TRANSACTION_PUTATTRIBUTETYPE
_TRANSACTION_PUTATTRIBUTETYPE_RES.fields_by_name['attributeType'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_PUTATTRIBUTETYPE_RES.containing_type = _TRANSACTION_PUTATTRIBUTETYPE
_TRANSACTION_PUTATTRIBUTETYPE.containing_type = _TRANSACTION
_TRANSACTION_PUTRELATIONTYPE_REQ.containing_type = _TRANSACTION_PUTRELATIONTYPE
_TRANSACTION_PUTRELATIONTYPE_RES.fields_by_name['relationType'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_PUTRELATIONTYPE_RES.containing_type = _TRANSACTION_PUTRELATIONTYPE
_TRANSACTION_PUTRELATIONTYPE.containing_type = _TRANSACTION
_TRANSACTION_PUTROLE_REQ.containing_type = _TRANSACTION_PUTROLE
_TRANSACTION_PUTROLE_RES.fields_by_name['role'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_PUTROLE_RES.containing_type = _TRANSACTION_PUTROLE
_TRANSACTION_PUTROLE.containing_type = _TRANSACTION
_TRANSACTION_PUTRULE_REQ.containing_type = _TRANSACTION_PUTRULE
_TRANSACTION_PUTRULE_RES.fields_by_name['rule'].message_type = session_dot_Concept__pb2._CONCEPT
_TRANSACTION_PUTRULE_RES.containing_type = _TRANSACTION_PUTRULE
_TRANSACTION_PUTRULE.containing_type = _TRANSACTION
_TRANSACTION_CONCEPTMETHOD_REQ.fields_by_name['method'].message_type = session_dot_Concept__pb2._METHOD_REQ
_TRANSACTION_CONCEPTMETHOD_REQ.containing_type = _TRANSACTION_CONCEPTMETHOD
_TRANSACTION_CONCEPTMETHOD_RES.fields_by_name['response'].message_type = session_dot_Concept__pb2._METHOD_RES
_TRANSACTION_CONCEPTMETHOD_RES.containing_type = _TRANSACTION_CONCEPTMETHOD
_TRANSACTION_CONCEPTMETHOD.containing_type = _TRANSACTION
_TRANSACTION_TYPE.containing_type = _TRANSACTION
DESCRIPTOR.message_types_by_name['Session'] = _SESSION
DESCRIPTOR.message_types_by_name['Transaction'] = _TRANSACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Session = _reflection.GeneratedProtocolMessageType('Session', (_message.Message,), dict(

  Open = _reflection.GeneratedProtocolMessageType('Open', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(

      MetadataEntry = _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), dict(
        DESCRIPTOR = _SESSION_OPEN_REQ_METADATAENTRY,
        __module__ = 'session.Session_pb2'
        # @@protoc_insertion_point(class_scope:session.Session.Open.Req.MetadataEntry)
        ))
      ,
      DESCRIPTOR = _SESSION_OPEN_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Session.Open.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _SESSION_OPEN_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Session.Open.Res)
      ))
    ,
    DESCRIPTOR = _SESSION_OPEN,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Session.Open)
    ))
  ,

  Close = _reflection.GeneratedProtocolMessageType('Close', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(

      MetadataEntry = _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), dict(
        DESCRIPTOR = _SESSION_CLOSE_REQ_METADATAENTRY,
        __module__ = 'session.Session_pb2'
        # @@protoc_insertion_point(class_scope:session.Session.Close.Req.MetadataEntry)
        ))
      ,
      DESCRIPTOR = _SESSION_CLOSE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Session.Close.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _SESSION_CLOSE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Session.Close.Res)
      ))
    ,
    DESCRIPTOR = _SESSION_CLOSE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Session.Close)
    ))
  ,
  DESCRIPTOR = _SESSION,
  __module__ = 'session.Session_pb2'
  # @@protoc_insertion_point(class_scope:session.Session)
  ))
_sym_db.RegisterMessage(Session)
_sym_db.RegisterMessage(Session.Open)
_sym_db.RegisterMessage(Session.Open.Req)
_sym_db.RegisterMessage(Session.Open.Req.MetadataEntry)
_sym_db.RegisterMessage(Session.Open.Res)
_sym_db.RegisterMessage(Session.Close)
_sym_db.RegisterMessage(Session.Close.Req)
_sym_db.RegisterMessage(Session.Close.Req.MetadataEntry)
_sym_db.RegisterMessage(Session.Close.Res)

Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), dict(

  Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(

    MetadataEntry = _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_REQ_METADATAENTRY,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Req.MetadataEntry)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_REQ,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Req)
    ))
  ,

  Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_RES,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Res)
    ))
  ,

  Iter = _reflection.GeneratedProtocolMessageType('Iter', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_ITER_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Iter.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_ITER_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Iter.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_ITER,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Iter)
    ))
  ,

  Open = _reflection.GeneratedProtocolMessageType('Open', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_OPEN_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Open.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_OPEN_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Open.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_OPEN,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Open)
    ))
  ,

  Commit = _reflection.GeneratedProtocolMessageType('Commit', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_COMMIT_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Commit.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_COMMIT_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Commit.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_COMMIT,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Commit)
    ))
  ,

  Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_QUERY_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Query.Req)
      ))
    ,

    Iter = _reflection.GeneratedProtocolMessageType('Iter', (_message.Message,), dict(

      Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
        DESCRIPTOR = _TRANSACTION_QUERY_ITER_RES,
        __module__ = 'session.Session_pb2'
        # @@protoc_insertion_point(class_scope:session.Transaction.Query.Iter.Res)
        ))
      ,
      DESCRIPTOR = _TRANSACTION_QUERY_ITER,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.Query.Iter)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_QUERY,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.Query)
    ))
  ,

  GetSchemaConcept = _reflection.GeneratedProtocolMessageType('GetSchemaConcept', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_GETSCHEMACONCEPT_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetSchemaConcept.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_GETSCHEMACONCEPT_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetSchemaConcept.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_GETSCHEMACONCEPT,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.GetSchemaConcept)
    ))
  ,

  GetConcept = _reflection.GeneratedProtocolMessageType('GetConcept', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_GETCONCEPT_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetConcept.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_GETCONCEPT_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetConcept.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_GETCONCEPT,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.GetConcept)
    ))
  ,

  GetAttributes = _reflection.GeneratedProtocolMessageType('GetAttributes', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_GETATTRIBUTES_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetAttributes.Req)
      ))
    ,

    Iter = _reflection.GeneratedProtocolMessageType('Iter', (_message.Message,), dict(

      Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
        DESCRIPTOR = _TRANSACTION_GETATTRIBUTES_ITER_RES,
        __module__ = 'session.Session_pb2'
        # @@protoc_insertion_point(class_scope:session.Transaction.GetAttributes.Iter.Res)
        ))
      ,
      DESCRIPTOR = _TRANSACTION_GETATTRIBUTES_ITER,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.GetAttributes.Iter)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_GETATTRIBUTES,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.GetAttributes)
    ))
  ,

  PutEntityType = _reflection.GeneratedProtocolMessageType('PutEntityType', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTENTITYTYPE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutEntityType.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTENTITYTYPE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutEntityType.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PUTENTITYTYPE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.PutEntityType)
    ))
  ,

  PutAttributeType = _reflection.GeneratedProtocolMessageType('PutAttributeType', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTATTRIBUTETYPE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutAttributeType.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTATTRIBUTETYPE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutAttributeType.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PUTATTRIBUTETYPE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.PutAttributeType)
    ))
  ,

  PutRelationType = _reflection.GeneratedProtocolMessageType('PutRelationType', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTRELATIONTYPE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRelationType.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTRELATIONTYPE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRelationType.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PUTRELATIONTYPE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.PutRelationType)
    ))
  ,

  PutRole = _reflection.GeneratedProtocolMessageType('PutRole', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTROLE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRole.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTROLE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRole.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PUTROLE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.PutRole)
    ))
  ,

  PutRule = _reflection.GeneratedProtocolMessageType('PutRule', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTRULE_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRule.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PUTRULE_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.PutRule.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PUTRULE,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.PutRule)
    ))
  ,

  ConceptMethod = _reflection.GeneratedProtocolMessageType('ConceptMethod', (_message.Message,), dict(

    Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_CONCEPTMETHOD_REQ,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.ConceptMethod.Req)
      ))
    ,

    Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_CONCEPTMETHOD_RES,
      __module__ = 'session.Session_pb2'
      # @@protoc_insertion_point(class_scope:session.Transaction.ConceptMethod.Res)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_CONCEPTMETHOD,
    __module__ = 'session.Session_pb2'
    # @@protoc_insertion_point(class_scope:session.Transaction.ConceptMethod)
    ))
  ,
  DESCRIPTOR = _TRANSACTION,
  __module__ = 'session.Session_pb2'
  # @@protoc_insertion_point(class_scope:session.Transaction)
  ))
_sym_db.RegisterMessage(Transaction)
_sym_db.RegisterMessage(Transaction.Req)
_sym_db.RegisterMessage(Transaction.Req.MetadataEntry)
_sym_db.RegisterMessage(Transaction.Res)
_sym_db.RegisterMessage(Transaction.Iter)
_sym_db.RegisterMessage(Transaction.Iter.Req)
_sym_db.RegisterMessage(Transaction.Iter.Res)
_sym_db.RegisterMessage(Transaction.Open)
_sym_db.RegisterMessage(Transaction.Open.Req)
_sym_db.RegisterMessage(Transaction.Open.Res)
_sym_db.RegisterMessage(Transaction.Commit)
_sym_db.RegisterMessage(Transaction.Commit.Req)
_sym_db.RegisterMessage(Transaction.Commit.Res)
_sym_db.RegisterMessage(Transaction.Query)
_sym_db.RegisterMessage(Transaction.Query.Req)
_sym_db.RegisterMessage(Transaction.Query.Iter)
_sym_db.RegisterMessage(Transaction.Query.Iter.Res)
_sym_db.RegisterMessage(Transaction.GetSchemaConcept)
_sym_db.RegisterMessage(Transaction.GetSchemaConcept.Req)
_sym_db.RegisterMessage(Transaction.GetSchemaConcept.Res)
_sym_db.RegisterMessage(Transaction.GetConcept)
_sym_db.RegisterMessage(Transaction.GetConcept.Req)
_sym_db.RegisterMessage(Transaction.GetConcept.Res)
_sym_db.RegisterMessage(Transaction.GetAttributes)
_sym_db.RegisterMessage(Transaction.GetAttributes.Req)
_sym_db.RegisterMessage(Transaction.GetAttributes.Iter)
_sym_db.RegisterMessage(Transaction.GetAttributes.Iter.Res)
_sym_db.RegisterMessage(Transaction.PutEntityType)
_sym_db.RegisterMessage(Transaction.PutEntityType.Req)
_sym_db.RegisterMessage(Transaction.PutEntityType.Res)
_sym_db.RegisterMessage(Transaction.PutAttributeType)
_sym_db.RegisterMessage(Transaction.PutAttributeType.Req)
_sym_db.RegisterMessage(Transaction.PutAttributeType.Res)
_sym_db.RegisterMessage(Transaction.PutRelationType)
_sym_db.RegisterMessage(Transaction.PutRelationType.Req)
_sym_db.RegisterMessage(Transaction.PutRelationType.Res)
_sym_db.RegisterMessage(Transaction.PutRole)
_sym_db.RegisterMessage(Transaction.PutRole.Req)
_sym_db.RegisterMessage(Transaction.PutRole.Res)
_sym_db.RegisterMessage(Transaction.PutRule)
_sym_db.RegisterMessage(Transaction.PutRule.Req)
_sym_db.RegisterMessage(Transaction.PutRule.Res)
_sym_db.RegisterMessage(Transaction.ConceptMethod)
_sym_db.RegisterMessage(Transaction.ConceptMethod.Req)
_sym_db.RegisterMessage(Transaction.ConceptMethod.Res)


DESCRIPTOR._options = None
_SESSION_OPEN_REQ_METADATAENTRY._options = None
_SESSION_CLOSE_REQ_METADATAENTRY._options = None
_TRANSACTION_REQ_METADATAENTRY._options = None

_SESSIONSERVICE = _descriptor.ServiceDescriptor(
  name='SessionService',
  full_name='session.SessionService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=4772,
  serialized_end=4986,
  methods=[
  _descriptor.MethodDescriptor(
    name='open',
    full_name='session.SessionService.open',
    index=0,
    containing_service=None,
    input_type=_SESSION_OPEN_REQ,
    output_type=_SESSION_OPEN_RES,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='close',
    full_name='session.SessionService.close',
    index=1,
    containing_service=None,
    input_type=_SESSION_CLOSE_REQ,
    output_type=_SESSION_CLOSE_RES,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='transaction',
    full_name='session.SessionService.transaction',
    index=2,
    containing_service=None,
    input_type=_TRANSACTION_REQ,
    output_type=_TRANSACTION_RES,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_SESSIONSERVICE)

DESCRIPTOR.services_by_name['SessionService'] = _SESSIONSERVICE

# @@protoc_insertion_point(module_scope)



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import numpy as np
from graph_nets import utils_np

from scipy.special import softmax


def compute_accuracy(target, output, use_nodes=True, use_edges=True):
    """Calculate model accuracy.

    Returns the number of elements correctly predicted to exist, and the number of completely correct graphs
    (100% correct predictions).

    Args:
        target: A `graphs.GraphsTuple` that contains the target graph.
        output: A `graphs.GraphsTuple` that contains the output graph.
        use_nodes: A `bool` indicator of whether to compute node accuracy or not.
        use_edges: A `bool` indicator of whether to compute edge accuracy or not.

    Returns:
        correct: A `float` fraction of correctly labeled nodes/edges.
        solved: A `float` fraction of graphs that are completely correctly labeled.

    Raises:
        ValueError: Nodes or edges (or both) must be used
    """
    if not use_nodes and not use_edges:
        raise ValueError("Nodes or edges (or both) must be used")
    tdds = utils_np.graphs_tuple_to_data_dicts(target)
    odds = utils_np.graphs_tuple_to_data_dicts(output)
    cs = []
    ss = []
    for td, od in zip(tdds, odds):
        xn = np.argmax(td["nodes"], axis=-1)
        yn = np.argmax(od["nodes"], axis=-1)
        xe = np.argmax(td["edges"], axis=-1)
        ye = np.argmax(od["edges"], axis=-1)
        c = []
        if use_nodes:
            c.append(xn == yn)
        if use_edges:
            c.append(xe == ye)
        c = np.concatenate(c, axis=0)
        s = np.all(c)
        cs.append(c)
        ss.append(s)
    correct = np.mean(np.concatenate(cs, axis=0))
    solved = np.mean(np.stack(ss))
    return correct, solved


def existence_accuracy(target, output, use_nodes=True, use_edges=True):
    if not use_nodes and not use_edges:
        raise ValueError("Nodes or edges (or both) must be used")
    tdds = utils_np.graphs_tuple_to_data_dicts(target)
    odds = utils_np.graphs_tuple_to_data_dicts(output)
    cs = []
    ss = []
    for td, od in zip(tdds, odds):

        nodes_to_predict = td["nodes"][:, 0] == 0
        xn = np.argmax(td["nodes"][:, 1:], axis=-1)
        xn = xn[nodes_to_predict]
        yn = np.argmax(softmax(od["nodes"][:, 1:], axis=1), axis=-1)
        yn = yn[nodes_to_predict]

        edges_to_predict = td["edges"][:, 0] == 0
        xe = np.argmax(td["edges"][:, 1:], axis=-1)
        xe = xe[edges_to_predict]
        ye = np.argmax(softmax(od["edges"][:, 1:], axis=1), axis=-1)
        ye = ye[edges_to_predict]

        c = []
        if use_nodes:
            c.append(xn == yn)
        if use_edges:
            c.append(xe == ye)
        c = np.concatenate(c, axis=0)
        s = np.all(c)
        cs.append(c)
        ss.append(s)
    correct = np.mean(np.concatenate(cs, axis=0))
    solved = np.mean(np.stack(ss))
    return correct, solved



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import abc

import sonnet as snt
import tensorflow as tf


class Attribute(snt.AbstractModule, abc.ABC):
    """
    Abstract base class for Attribute value embedding models
    """
    def __init__(self, attr_embedding_dim, name='AttributeEmbedder'):
        super(Attribute, self).__init__(name=name)
        self._attr_embedding_dim = attr_embedding_dim


class ContinuousAttribute(Attribute):
    def __init__(self, attr_embedding_dim, name='ContinuousAttributeEmbedder'):
        super(ContinuousAttribute, self).__init__(attr_embedding_dim, name=name)

    def _build(self, attribute_value):
        tf.summary.histogram('cont_attribute_value_histogram', attribute_value)
        embedding = snt.Sequential([
            snt.nets.MLP([self._attr_embedding_dim] * 3, activate_final=True, use_dropout=True),
            snt.LayerNorm(),
        ])(tf.cast(attribute_value, dtype=tf.float32))
        tf.summary.histogram('cont_embedding_histogram', embedding)
        return embedding


class CategoricalAttribute(Attribute):
    def __init__(self, num_categories, attr_embedding_dim, name='CategoricalAttributeEmbedder'):
        super(CategoricalAttribute, self).__init__(attr_embedding_dim, name=name)

        self._num_categories = num_categories

    def _build(self, attribute_value):
        int_attribute_value = tf.cast(attribute_value, dtype=tf.int32)
        tf.summary.histogram('cat_attribute_value_histogram', int_attribute_value)
        embedding = snt.Embed(self._num_categories, self._attr_embedding_dim)(int_attribute_value)
        tf.summary.histogram('cat_embedding_histogram', embedding)
        return tf.squeeze(embedding, axis=1)


class BlankAttribute(Attribute):

    def __init__(self, attr_embedding_dim, name='BlankAttributeEmbedder'):
        super(BlankAttribute, self).__init__(attr_embedding_dim, name=name)

    def _build(self, attribute_value):
        shape = tf.stack([tf.shape(attribute_value)[0], self._attr_embedding_dim])

        encoded_features = tf.zeros(shape, dtype=tf.float32)
        return encoded_features



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.session.Concept_pb2 as ConceptMessages


# base type constant names 
CONCEPTS = META_TYPE, ATTRIBUTE_TYPE, RELATION_TYPE, ENTITY_TYPE, ENTITY, ATTRIBUTE, RELATION, ROLE, RULE = \
    "META_TYPE", "ATTRIBUTE_TYPE", "RELATION_TYPE", "ENTITY_TYPE", "ENTITY", "ATTRIBUTE", "RELATION", "ROLE", "RULE"

"""
NOTE: the string META_TYPE is the name of the programmatic type of
Thing, Entity, Attribute, Relation IN GRPC. In the original Java server implementation,
these have the type TYPE, but due to bad naming at some point, GRPC
says the base_type of them is META_TYPE. Thus, there will never be
any concepts with base_type TYPE, only META_TYPE on GRPC-connected clients.
To match the server class hierarchy, I here instantiate TYPE objects rather than
META_TYPE, and when the time comes we will rename META_TYPE to TYPE on GRPC connected
clients too.
"""

grpc_base_types = ConceptMessages.Concept.BASE_TYPE
grpc_base_type_to_name = {
    grpc_base_types.Value("META_TYPE"): META_TYPE,
    grpc_base_types.Value("ENTITY_TYPE"): ENTITY_TYPE,
    grpc_base_types.Value("RELATION_TYPE"): RELATION_TYPE,
    grpc_base_types.Value("ATTRIBUTE_TYPE"): ATTRIBUTE_TYPE,
    grpc_base_types.Value("ROLE"): ROLE,
    grpc_base_types.Value("RULE"): RULE,
    grpc_base_types.Value("ENTITY"): ENTITY,
    grpc_base_types.Value("RELATION"): RELATION,
    grpc_base_types.Value("ATTRIBUTE"): ATTRIBUTE 
}

# reverse lookup of above
# note: assuming one-to-one correspondence
name_to_grpc_base_type = dict(zip(grpc_base_type_to_name.values(), grpc_base_type_to_name.keys()))






#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

from graph_nets import utils_tf, utils_np


def create_placeholders(input_graphs, target_graphs):
    """
    Creates placeholders for the model training and evaluation.
    Returns:
    input_ph: The input graph's placeholders, as a graph namedtuple.
    target_ph: The target graph's placeholders, as a graph namedtuple.
    """
    input_ph = utils_tf.placeholders_from_networkxs(input_graphs, name="input_placeholders_from_networksx")
    target_ph = utils_tf.placeholders_from_networkxs(target_graphs, name="target_placeholders_from_networkxs")
    return input_ph, target_ph


def create_feed_dict(input_ph, target_ph, inputs, targets):
    """Creates the feed dict for the placeholders for the model training and evaluation.

    Args:
        input_ph: The input graph's placeholders, as a graph namedtuple.
        target_ph: The target graph's placeholders, as a graph namedtuple.
        inputs: The input graphs
        targets: The target graphs

    Returns:
        feed_dict: The feed `dict` of input and target placeholders and data.
    """
    input_graphs = utils_np.networkxs_to_graphs_tuple(inputs)
    target_graphs = utils_np.networkxs_to_graphs_tuple(targets)
    feed_dict = {input_ph: input_graphs, target_ph: target_graphs}
    return feed_dict


def make_all_runnable_in_session(*args):
    """Lets an iterable of TF graphs be output from a session as NP graphs."""
    return [utils_tf.make_runnable_in_session(a) for a in args]



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#


class PropertyComparable:
    """
    Methods to give to an object such that it can be compared with another object based on their
    properties/attributes. Avoided using the name 'Attribute' since this is common Grakn terminology.
    """
    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        """Overrides the default implementation"""
        return hash(tuple(sorted(self.__dict__.items())))
